'''
Utility functions for "Data Mining for Business Analytics: Concepts, Techniques, and 
Applications in Python"

(c) 2019 Galit Shmueli, Peter C. Bruce, Peter Gedeck 
'''
import unittest

import pytest

from mistat.simulation.powerCircuitSimulation import PowerCircuitSimulation


class TestPowerCircuitSimulation(unittest.TestCase):

    def test_PowerCircuitSimulation(self):
        simulator = PowerCircuitSimulation(seed=1)
        result = simulator.simulate()
        assert result.shape == (50, 14)
        assert result['volts'].mean() == pytest.approx(229.610782)
        assert result['volts'].std() == pytest.approx(3.495034)



'''
Created on Jun 22, 2020

@author: petergedeck
'''
import unittest

from mistat.qcc.ewmaChart import ewmaSmooth
import numpy as np


class TestCusum(unittest.TestCase):
    def test_ewmaSmooth(self):
        smoothed = ewmaSmooth([1, 2, 3, 4])
        np.testing.assert_array_almost_equal(smoothed['y'], [1.000, 1.200, 1.560, 2.048])

        smoothed = ewmaSmooth([1, 2, 3, 4], x=[2, 3, 4, 1])
        np.testing.assert_array_almost_equal(smoothed['x'], [1, 2, 3, 4])
        np.testing.assert_array_almost_equal(smoothed['y'], [4, 3.4, 3.12, 3.096])

        smoothed = ewmaSmooth([1, 2, 3, 4], smooth=0.8)
        np.testing.assert_array_almost_equal(smoothed['y'], [1, 1.8, 2.76, 3.752])

        smoothed = ewmaSmooth([1, 2, 3, 4], start=10)
        np.testing.assert_array_almost_equal(smoothed['y'], [8.2, 6.96, 6.168, 5.7344])



from .bandit import simulateOAB, optimalOAB
from .dodge.dodge_sequential import sequentialDesign
from .dodge.dodge_single import SSPlanBinomial, SSPlanHyper, SSPlanPoisson
from .generic import findPlan, findPlanApprox
from .oc import OperatingCharacteristics2c



'''
Created on Jun 20, 2020

@author: petergedeck
'''
from numbers import Number

from scipy import stats

from mistat.qcc.rules import shewhartRules
from mistat.qcc.statistics import qccStatistics, GroupMeans
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


# qcc <- function(data, type = c("xbar", "R", "S", "xbar.one", "p", "np", "c", "u", "g"), sizes,
#                 center, std.dev, limits, data.name, labels, newdata, newsizes, newdata.name,
#                 newlabels, nsigmas = 3, confidence.level, rules = shewhart.rules, plot = TRUE, ...)
class QualityControlChart:
    def __init__(self, data, qcc_type=qccStatistics.default, labels=None,
                 center=None, std_dev=None, limits=None, sizes=None,
                 nsigmas=3, confidence_level=None,
                 newdata=None):
        self.statistic = qccStatistics.get(qcc_type)
        self.qcc_type = self.statistic.qcc_type

        self.data = np.array(data)
        if len(self.data.shape) == 1:
            self.data = pd.DataFrame({'data': self.data})
        self.sizes = sizes
        if self.sizes is None:
            self.sizes = self.statistic.getSizes(data)

        if labels is None:
            labels = list(range(len(data)))
        self.labels = labels

        if isinstance(self.sizes, Number):
            self.stats = GroupMeans((self.data.values / self.sizes).flatten(), np.mean(self.data))
            self.sizes = [self.sizes] * len(self.data)
        else:
            self.stats = self.statistic.stats(self.data, self.sizes)
        self.center = center or self.stats.center

        self.std_dev = self.statistic.sd(self.data, std_dev, self.sizes)

        if confidence_level is None:
            self.nsigmas = nsigmas
            conf = nsigmas
        else:
            self.confidence_level = confidence_level
            conf = confidence_level
        if limits is None:
            self.limits = self.statistic.limits(self.center, self.std_dev, self.sizes, conf)
        else:
            self.limits = limits

        if newdata is not None:
            pass
#             raise NotImplementedError()
            # gives self.newdata, self.newstats, self.newsizes
        else:
            self.newdata = None
            self.newstats = None
            self.newsizes = None

        # TODO violations
        self.violations = shewhartRules(self)

    def plot(self, title=None, ax=None):
        if ax is None:
            _, ax = plt.subplots(figsize=(8, 6))
        beyondLimits = [*self.violations['beyondLimits']['LCL'], *self.violations['beyondLimits']['UCL']]
        violatingRuns = self.violations['violatingRuns']
        df = pd.DataFrame({'x': self.labels, 'y': self.stats.statistics})
        ax = df.plot.line(x='x', y='y', style='-o', color='lightgrey',
                          marker='o', markerfacecolor='black', ax=ax)
        ax.plot(df.iloc[beyondLimits]['x'], df.iloc[beyondLimits]['y'], linestyle='None',
                marker='s', markerfacecolor='red', markeredgecolor='red')
        ax.plot(df.iloc[violatingRuns]['x'], df.iloc[violatingRuns]['y'],
                linestyle='None', marker='s', markerfacecolor='red', markeredgecolor='red')
        ax.legend().remove()
        ax.set_xlabel('Group')
        ax.set_ylabel('Group summary statistics')

        ax.axhline(self.center, color='black')
        secax = ax.secondary_yaxis('right')
        secax.yaxis.tick_right()
        if len(self.limits) == 1:
            ax.axhline(self.limits.LCL[0], color='black', linestyle='--')
            ax.axhline(self.limits.UCL[0], color='black', linestyle='--')
            ax.set_yticks([self.limits.LCL[0], self.limits.UCL[0], self.center])
            ax.set_yticklabels(['LCL', 'UCL', 'CL'])

        fig = ax.get_figure()
        fig.subplots_adjust(bottom=0.2)
        fig.text(0.1, 0.1, f'Number of groups = {len(self.labels)}', fontsize=12)
        fig.text(0.1, 0.06, f'Center = {self.center:.5g}', fontsize=12)
        fig.text(0.1, 0.02, f'StdDev = {self.std_dev:.5g}', fontsize=12)
        if len(self.limits) == 1:
            fig.text(0.4, 0.06, f'LCL = {self.limits.LCL[0]:.5g}', fontsize=12)
            fig.text(0.4, 0.02, f'UCL = {self.limits.UCL[0]:.5g}', fontsize=12)
        fig.text(0.6, 0.06, f'Number beyond limits = {len(beyondLimits)}', fontsize=12)
        fig.text(0.6, 0.02, f'Number violating runs = {len(violatingRuns)}', fontsize=12)

        fig_title = [f'{self.statistic.qcc_type} Chart']
        if title is not None:
            fig_title.append(title)
        fig.suptitle('\n'.join(fig_title), fontsize=14)
        return ax

    def oc_curves(self, nsigmas=None, ax=None):
        if self.qcc_type in ['p' or 'np']:
            return oc_curves_p(self, ax=ax)
        else:
            raise NotImplementedError()


def oc_curves_p(qcc, ax=None):
    if qcc.qcc_type not in ['p', 'np']:
        raise ValueError("not a qcc object of type 'p', or 'np'.")

    size = set(qcc.sizes)
    if len(size) > 1:
        raise ValueError('Operating characteristic curves available only for equal sample sizes!')
    if qcc.limits is None:
        raise ValueError("the `qcc' object does not have control limits!")

    if ax is None:
        _, ax = plt.subplots(figsize=(8, 6))

    limits = qcc.limits
    size = qcc.sizes[0]
    p = np.linspace(0, 1, 101)

    if qcc.qcc_type == 'p':
        UCL = min(np.floor(size * limits.UCL[0]), size)
        LCL = max(np.floor(size * limits.LCL[0]), 0)
    else:
        UCL = min(np.floor(limits.UCL[0]), size)
        LCL = max(np.floor(limits.LCL[0]), 0)
    beta = pd.Series(stats.binom(size, p).cdf(UCL) - stats.binom(size, p).cdf(LCL - 1),
                     index=p)

    beta.plot(ax=ax, color='black')
    ax.set_xlim(-0.05, 1.05)
    ax.set_ylim(-0.05, 1.05)
    ax.set_xlabel(r'$p$')
    ax.set_ylabel('Prob. type II error')
    ax.set_title(f'OC curves for {qcc.qcc_type} Chart')

    p_max = beta.idxmax()
    ax.plot([p_max, p_max], [0, beta.max()], color='grey', linestyle='--')


def qcc_groups(data, groups):
    grouped = pd.Series(data).groupby(groups)

    max_size = max(grouped.size())

    result = np.empty((len(grouped), max_size))
    result[:] = np.NaN
    for idx, group in grouped:
        result[idx - 1, :len(group.values)] = group.values
    return result


def qcc_overdispersion_test(x, sizes=None, dist=None):
    if dist is None:
        dist = 'poisson' if sizes is None else 'binomial'
    dist = dist.lower()
    if dist == 'binomial' and sizes is None:
        raise ValueError("binomial data require argument 'size'")
    if sizes is not None and len(x) != len(sizes):
        raise ValueError("arguments 'x' and 'sizes' must have same length")

    if dist == 'binomial':
        p = np.sum(x) / np.sum(sizes)
        theor_var = np.mean(sizes) * p * (1 - p)
    elif dist == 'poisson':
        theor_var = np.mean(x)
    else:
        raise ValueError(f"invalid distribution '{dist}'")
    n = len(x)
    obs_var = np.var(x, ddof=1)

    D = (n - 1) * obs_var / theor_var
    p_value = 1 - stats.chi2(n - 1).cdf(D)

    return pd.Series({
        'Overdispersion test': dist,
        'Obs.Var/Theor.Var': obs_var / theor_var,
        'Statistic': D,
        'p-value': p_value,
    })



'''
Created on Jun 13, 2020

@author: petergedeck
'''
from dataclasses import dataclass
import pandas as pd


@dataclass
class MistatSimulation:
    def simulate(self):
        raise NotImplementedError('Simulation must implement method simulate')


class SimulationResult(pd.DataFrame):
    pass


def simulationGroup(simulation_result, group_size, quiet=False):
    if not isinstance(simulation_result, SimulationResult):
        raise ValueError('Simulator must be of type MistatSimulation')

    ngroups = len(simulation_result) // group_size
    group_labels = list(repeat_elements(list(range(1, ngroups + 1)), group_size))

    if len(group_labels) < len(simulation_result):
        ndiscard = len(simulation_result) - len(group_labels)
        if not quiet:
            print(f'discarded first {ndiscard} observations because the number of '
                  'observations is not a multiple of group_size')
        simulation_result = simulation_result.iloc[ndiscard:, :].copy()
    simulation_result['group'] = group_labels
    return simulation_result


def convert_to_list(value):
    if isinstance(value, (list, tuple)):
        return value
    return [value]


def repeat_elements(values, nrepeat):
    for value in convert_to_list(values):
        yield from [value] * nrepeat



'''
Created on Jun 12, 2020

@author: gedeck
'''
from pathlib import Path
import string

import pandas as pd


DATA_DIR = Path(__file__).parent / 'csvFiles'


def load_data(name):
    """ Returns the data either as a Pandas data frame or series """
    data_file = get_data_file(name)
    if not data_file.exists():
        raise ValueError('Data file {name} not found')
    data = pd.read_csv(data_file)
    if data.shape[1] == 1:
        return data[data.columns[0]]
    return data


def describe_data(name):
    """ Return information about the data file """
    description_file = get_description_file(name)
    if not description_file.exists():
        raise ValueError('Description for data file {name} not found')
    text = description_file.read_text()
    text = text.replace('R Documentation', 'Documentation')
    return text


def get_data_file(name):
    if name.endswith('.gz'):
        name = name[:-3]
    if name.endswith('.csv'):
        name = name[:-4]
    return DATA_DIR / f'{name}.csv.gz'


def get_description_file(name):
    if name.endswith('.gz'):
        name = name[:-3]
    if name.endswith('.csv'):
        name = name[:-4]
    if name.endswith('.md'):
        name = name[:-3]
    description_file = DATA_DIR / 'md' / f'{name}.md'
    if not description_file.exists():
        name = description_file.with_suffix('').name.rstrip(string.digits)
        description_file = DATA_DIR / 'md' / f'{name}.md'
    return description_file



'''
Created on Jun 13, 2020

@author: petergedeck
'''
from dataclasses import dataclass, fields

from scipy import stats

from mistat.simulation.mistatSimulation import SimulationResult
import numpy as np
import pandas as pd

from .mistatSimulation import MistatSimulation, repeat_elements, convert_to_list


errors = {
    'm': 0.1,
    's': 0.01,
    'v0': 0.00025,
    'k': 50,
    'p0': 0.01,
    't': 0.13,
    't0': 0.13
}


@dataclass
class PowerCircuitSimulation(MistatSimulation):
    rsA: float = 8_200
    rsB: float = 220_000
    rsC: float = 1_000
    rsD: float = 33_000
    rsE: float = 56_000
    rsF: float = 5_600
    rsG: float = 3_300
    rsH: float = 58.5
    rsI: float = 1_000
    rsJ: float = 120
    trK: float = 130
    trL: float = 100
    trM: float = 130
    tlA: float = 5
    tlB: float = 10
    tlC: float = 10
    tlD: float = 5
    tlE: float = 5
    tlF: float = 5
    tlG: float = 10
    tlH: float = 5
    tlI: float = 5
    tlJ: float = 5
    tlK: float = 5
    tlL: float = 10
    tlM: float = 5

    n_simulation: int = 50  # desired number of simulations
    seed: float = None
    check: bool = True

    def __post_init__(self):
        if self.seed is not None:
            np.random.seed(seed=self.seed)

        self.resistors = [field.name for field in fields(self)
                          if field.name.startswith('rs')]
        self.transistors = [field.name for field in fields(self)
                            if field.name.startswith('tr')]
        self.tolerances = [field.name for field in fields(self)
                           if field.name.startswith('tl')]
        self.options = [*self.resistors, *self.transistors, *self.tolerances]

        for option in self.options:
            validate_range(getattr(self, option), 0, f'{option} must be greater than 0')

        if self.n_simulation < 1:
            raise ValueError('Number of simulations must be greater 1')

        # Convert to lists
        maxsize = 0
        for option in self.options:
            values = convert_to_list(getattr(self, option))
            maxsize = max(maxsize, len(values))
            setattr(self, option, values)
        if maxsize == 1:
            maxsize = self.n_simulation

        # Make sure that the vectors are all the same length
        for option in self.options:
            values = getattr(self, option)
            if maxsize % len(values) != 0:
                raise ValueError(f'Inconsistent length of option {option}')
            values = np.array(list(repeat_elements(values, maxsize // len(values))))
            setattr(self, option, values)

        self.df = pd.DataFrame({option: getattr(self, option) for option in self.options})

    def simulate(self):
        size = len(self.rsA)

        # add errors
        data = {}
        for option in self.options:
            if option.startswith('tl'):
                element = f'rs{option[2]}' if hasattr(self, f'rs{option[2]}') else f'tr{option[2]}'
                rv = stats.norm.rvs(size=size)
                rv = rv * 2 * getattr(self, option) * getattr(self, element) / 600
                data[element] = data[element] + rv
            else:
                data[option] = getattr(self, option)

        # calculate resulting voltage
        data['volts'] = self.vCircuit(**data)

        # copy specified values into results
        for option in self.resistors:
            data[option] = getattr(self, option)
        for option in self.transistors:
            data[option] = getattr(self, option)
        return SimulationResult(data)

    @staticmethod
    def vCircuit(rsA, rsB, rsC, rsD, rsE, rsF, rsG, rsH, rsI, rsJ, trK, trL, trM):
        a = rsB / (rsA + rsB)
        b = (rsA * rsB / (rsA + rsB) + rsC) / (trL * trM) + rsI
        c = rsE + 0.5 * rsG
        d = rsA * rsB * trK / (rsA + rsB)
        e = rsF + 0.5 * rsG
        f = (c + e) * (1 + trK) * rsH + c * e
        g = rsH + 0.6
        h = 1.2

        aden = 1 + d * e / f + b * (1 / rsJ + 0.006 * (1 + 13.67 / rsJ)) + a * 0.1367 * 0.6
        anum = (a + b / rsJ) * (138 - 1.33) + d * (c + e) * g / f - h

        return anum / aden


def validate_range(value, left, message):
    left = left - 1e-6
    if isinstance(value, (tuple, list, pd.core.series.Series, np.ndarray)):
        if not all(left <= v for v in value):
            raise ValueError(message)
    else:
        if not left <= value:
            raise ValueError(message)



'''
Created on Jan 10, 2021

@author: petergedeck
'''
import random
import numpy as np
import pingouin as pg
import pandas as pd


def availabilityEBD(ttf, ttr, n, seed=None, printSummary=True):
    ''' given TTF and TTR, determines the availbility EBD and calculates asymptotic availability '''
    if seed is not None:
        np.random.seed(seed)
    ttf = np.array(ttf)
    ttr = np.array(ttr)

    _, asf = pg.compute_bootci(ttf, func='mean', n_boot=1000, return_dist=True)
    _, asr = pg.compute_bootci(ttr, func='mean', n_boot=1000, return_dist=True)
    bootavail = pd.Series(asf / (asf + asr), name='availability EBD')
    if printSummary:
        avf = np.mean(asf)
        avr = np.mean(asr)
        avail = avf / (avf + avr)
        print(f'The estimated MTTF from ttf is {avf:.2f}')
        print(f'The estimated MTTR from ttr is {avr:.2f}')
        print(f'The estimated asymptotic availability is {avail:.4f}')
        print()
        print(bootavail.describe())
    return bootavail


def renewalEBD(ttf, ttr, time, n, seed=None, printSummary=True):
    ''' given TTF and TTR, determines the EBD of the number of renewals '''
    if seed is not None:
        random.seed(seed)
    ttf = np.array(ttf)
    ttr = np.array(ttr)

    result = []
    for _ in range(n):
        tt = 0
        it = 1
        while tt < time:
            tt += random.choice(ttf)
            tt += random.choice(ttr)
            it += 1
        result.append(it)

    result = pd.Series(result, name='number of renewals EBD')
    if printSummary:
        print(f'The estimated MEAN NUMBER Of RENEWALS is {np.mean(result):.2f}')
        print(result.describe())
    return result



'''
Utility functions for "Data Mining for Business Analytics: Concepts, Techniques, and 
Applications in Python"

(c) 2019 Galit Shmueli, Peter C. Bruce, Peter Gedeck 
'''
import unittest

import pytest

from mistat.data import load_data
from mistat.qcc.qualityControlChart import qcc_groups, QualityControlChart,\
    qcc_overdispersion_test
from mistat.simulation.mistatSimulation import simulationGroup
from mistat.simulation.pistonSimulation import PistonSimulator
import numpy as np


def test_qcc_groups(num_regression):
    Ps = PistonSimulator(seed=1).simulate()
    Ps = simulationGroup(Ps, 5)

    result = qcc_groups(Ps['seconds'], Ps['group'])
    num_regression.check({f'row_{i}': r for i, r in enumerate(result)})


def test_qcc_groups_missing_values():
    result = qcc_groups([1, 2, 2, 3, 3, 3, 4, 4, 4, 4], [1, 2, 2, 3, 3, 3, 4, 4, 4, 4])
    expect = np.array([[1.0, np.NaN, np.NaN, np.NaN],
                       [2.0, 2.0, np.NaN, np.NaN],
                       [3.0, 3.0, 3.0, np.NaN],
                       [4.0, 4.0, 4.0, 4.0]])
    np.testing.assert_array_equal(result, expect)


class TestQCCOverdispersionTest(unittest.TestCase):
    def test_exceptions(self):
        with pytest.raises(ValueError):
            qcc_overdispersion_test([1, 2, 3], dist='binomial')
        with pytest.raises(ValueError):
            qcc_overdispersion_test([1, 2, 3], sizes=[1, 2])
        with pytest.raises(ValueError):
            qcc_overdispersion_test([1, 2, 3], sizes=[1, 2, 3], dist='typo')

    def test_results(self):
        # data from Wetherill and Brown (1991) pp. 212--213, 216--218:
        x = [12, 11, 18, 11, 10, 16, 9, 11, 14, 15, 11, 9, 10, 13, 12,
             8, 12, 13, 10, 12, 13, 16, 12, 18, 16, 10, 16, 10, 12, 14]

        sizes = [50] * len(x)
        result = qcc_overdispersion_test(x, sizes)
        assert result['Overdispersion test'] == 'binomial'
        assert result['Obs.Var/Theor.Var'] == pytest.approx(0.7644566)
        assert result['Statistic'] == pytest.approx(22.16924)
        assert result['p-value'] == pytest.approx(0.8131149)

        x = [11, 8, 13, 11, 13, 17, 25, 23, 11, 16, 9, 15, 10, 16, 12,
             8, 9, 15, 4, 12, 12, 12, 15, 17, 14, 17, 12, 12, 7, 16]
        result = qcc_overdispersion_test(x)
        assert result['Overdispersion test'] == 'poisson'
        assert result['Obs.Var/Theor.Var'] == pytest.approx(1.472203)
        assert result['Statistic'] == pytest.approx(42.69388)
        assert result['p-value'] == pytest.approx(0.048579, rel=1e-5)


class TestQualityControlChart(unittest.TestCase):
    def setUp(self):
        Ps = PistonSimulator(seed=1).simulate()
        Ps = simulationGroup(Ps, 5)
        self.cycleTime = qcc_groups(Ps['seconds'], Ps['group'])

    def test_xbar_chart(self):
        _ = QualityControlChart(self.cycleTime)
        _ = QualityControlChart(self.cycleTime, qcc_type='xbar')
        _ = QualityControlChart(self.cycleTime, qcc_type='S')
        _ = QualityControlChart(load_data('JANDEFECT'), qcc_type='p', sizes=100,
                                center=0.048, std_dev=np.sqrt(0.048 * (1 - 0.048)))
        _ = QualityControlChart(load_data('CONTACTLEN'), qcc_type='R')

    def test_S_chart(self):
        _ = QualityControlChart(self.cycleTime, qcc_type='S')



'''
Created on Jun 26, 2020

@author: petergedeck
'''
import unittest

from mistat.acceptanceSampling.generic import findPlan, Plan, PlanNormal, find_k,\
    findPlanApprox
import numpy as np
import pandas as pd
from unittest.case import skip


class TestGeneric(unittest.TestCase):
    def test_findPlan(self):
        result = findPlan(PRP=(0.01, 0.05), CRP=(0.08, 0.05), oc_type="hypergeom", N=100)
        assert result == Plan(31, 0, 1)

        p0 = np.array([*[0.01] * 10, *[0.03] * 10])
        pt = np.linspace(0.05, 0.32, 10)
        pt = np.array([*pt, *pt])

        result = []
        for p0i, pti in zip(p0, pt):
            result.append([p0i, pti, *findPlan(PRP=(p0i, 0.05), CRP=(pti, 0.05), oc_type="hypergeom", N=100)])
        result = pd.DataFrame(result, columns=['p0', 'pt', 'n', 'c', 'r'])
        np.testing.assert_array_equal(result['n'], [45, 31, 23, 18, 15, 13, 11, 10, 9, 8] * 2)
        np.testing.assert_array_equal(result['c'], [0] * 20)

        result = []
        for p0i, pti in zip(p0, pt):
            result.append([p0i, pti, *findPlan(PRP=(p0i, 0.1), CRP=(pti, 0.2), oc_type="hypergeom", N=100)])
        result = pd.DataFrame(result, columns=['p0', 'pt', 'n', 'c', 'r'])
        np.testing.assert_array_equal(result['n'], [27, 18, 13, 11, 9, 7, 6, 6, 5, 5] * 2)
        np.testing.assert_array_equal(result['c'], [0] * 20)

        result = []
        for p0i, pti in zip(p0, pt):
            result.append([p0i, pti, *findPlan(PRP=(p0i, 0.5), CRP=(pti, 0.3), oc_type="hypergeom", N=100)])
        result = pd.DataFrame(result, columns=['p0', 'pt', 'n', 'c', 'r'])
        np.testing.assert_array_equal(result['n'], [21, 14, 10, 8, 7, 6, 5, 4, 4, 4, 42, 14, 10, 8, 7, 6, 5, 4, 4, 4])
        np.testing.assert_array_equal(result['c'], [*[0] * 10, 1, *[0] * 9])

    def test_findPlan_normal(self):
        result = findPlan(PRP=(0.01, 0.95), CRP=(0.03, 0.05), oc_type="normal", N=500)
        assert result == PlanNormal(55, 2.1045558562282713, 'known')

        result = findPlan(PRP=(0.01, 0.95), CRP=(0.03, 0.05), oc_type="normal", N=500, s_type='unknown')
        assert result == PlanNormal(176, 2.105480330251979, 'unknown')

    @skip
    def test_findPlan_poisson(self):
        result = findPlan(PRP=(0.01, 0.95), CRP=(0.03, 0.05), oc_type="poisson", N=500)
        assert result == Plan(524, 9, 10)

    def test_findPlanApprox(self):
        result = findPlanApprox(PRP=(0.01, 0.95), CRP=(0.03, 0.05),  N=500)
        assert result == Plan(248, 4, 5)
        result = findPlanApprox(PRP=(0.01, 0.95), CRP=(0.05, 0.05),  N=500)
        assert result == Plan(127, 2, 3)

    def test_find_k(self):
        assert find_k(100, 0.01, 0.95, [0, 1000]) == 2.0400552555161156
        assert find_k(10, 0.01, 0.95, [0, 1000]) == 1.562533782230915



'''
Created on Dec 30, 2020

@author: petergedeck
'''


from pathlib import Path
import unittest

from mistat.rsm.responseSurfaceMethod import ResponseSurfaceMethod
import pandas as pd
import numpy as np
import statsmodels.formula.api as smf
import pytest


class TestResponseSurfaceMethod(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.data1 = loadModel('simulationData1.csv')
        cls.data2 = loadModel('simulationData2.csv')

    def test_function(self):
        model = self.data1['model']
        codes = self.data1['codes']
        rsm = ResponseSurfaceMethod(model, codes)
        assert rsm.f([0, 0, 0, 0]) == pytest.approx(0.559, abs=0.001)
        assert rsm.f([1.260, -1.244, 0.870, 0.318]) == pytest.approx(0.415, abs=0.001)

        model = self.data2['model']
        codes = self.data2['codes']
        rsm = ResponseSurfaceMethod(model, codes)
        assert rsm.f([0, 0, 0, 0]) == pytest.approx(0.586, abs=0.001)
        assert rsm.f([1.808, -0.167, 0.449, -0.705]) == pytest.approx(0.406, abs=0.001)

    def test_stationary(self):
        model = self.data1['model']
        codes = self.data1['codes']
        rsm = ResponseSurfaceMethod(model, codes)
        np.testing.assert_allclose(rsm.stationary_point(), np.array(
            [-10.642838, 2.957658, 3.660820, -2.299632]), rtol=1e-4)

        model = self.data2['model']
        codes = self.data2['codes']
        rsm = ResponseSurfaceMethod(model, codes)
        np.testing.assert_allclose(rsm.stationary_point(), np.array(
            [-1.1969668, 0.9320400, 2.2297196, 0.4233004]), rtol=1e-4)

    def test_eigen(self):
        model = self.data1['model']
        codes = self.data1['codes']
        rsm = ResponseSurfaceMethod(model, codes)

        # first test the complex eigenvalues
        eigen = rsm._eigen()
        ev = sorted(eigen['eval'].real, reverse=True)
        np.testing.assert_allclose(ev, np.array(
            [0.034424442, 0.008661927, 0.003316576, -0.001943580]), rtol=1e-4)
        ev = sorted(ev.imag for ev in eigen['eval'])
        np.testing.assert_allclose(ev, np.zeros(4), rtol=1e-4)

        # nowtest the sorted real eigenvalues/vectors
        eigen = rsm.eigen()
        np.testing.assert_allclose(eigen['eval'], np.array(
            [0.034424442, 0.008661927, 0.003316576, -0.001943580]), rtol=1e-4)
        evec1 = eigen['evec'][:, 0]
        if evec1[0] < 0:
            evec1 = -evec1
        np.testing.assert_allclose(evec1, np.array(
            [0.2722108, 0.2597354, 0.9261707, 0.0254271]), rtol=1e-4)


def loadModel(filename):
    factors = {
        's': [0.01, 0.015],
        'v0': [0.00625, 0.00875],
        'k': [2000, 4000],
        't0': [345, 355],
    }
    result = pd.read_csv(Path(__file__).parent / filename)
    # calculate mean and std of response by group
    result = result.groupby(by='group')
    result = result.agg({'s': 'mean', 'v0': 'mean', 'k': 'mean',
                         't0': 'mean', 'seconds': ['mean', 'std']})
    result.columns = ['s', 'v0', 'k', 't0', 'Ymean', 'Ystd']

    # transformation between factors and code levels
    factor2x = {factor: f'x{i}' for i, factor in enumerate(factors, 1)}
    x2factor = {f'x{i}': factor for i, factor in enumerate(factors, 1)}
    center = {factor: 0.5 * (max(values) + min(values)) for factor, values in factors.items()}
    unit = {factor: 0.5 * (max(values) - min(values)) for factor, values in factors.items()}

    # define helper function to convert code co-ordinates to factor co-ordinates
    def toFactor(code, codeValue):
        ''' convert code to factor co-ordinates '''
        factor = x2factor[code]
        return center[factor] + codeValue * unit[factor]

    # add code levels to table
    for c in factors:
        result[factor2x[c]] = (result[c] - center[c]) / unit[c]
    formula = ('Ymean ~ (x1+x2+x3+x4)**2 + ' +
               'np.power(x1,2) + np.power(x2,2) + np.power(x3,2) + np.power(x4,2)')
    return {'codes': list(x2factor), 'model': smf.ols(formula, data=result).fit()}



'''
Created on Jan 10, 2021

@author: petergedeck
'''
import pandas as pd
import numpy as np


def plotECDF(data):
    n = len(data)
    ecdf = pd.DataFrame({
        'x': np.sort(data),
        'Fn(x)': range(1, n + 1),
    })
    ecdf['Fn(x)'] = ecdf['Fn(x)'] / n
    ax = ecdf.plot(x='x', y='Fn(x)', color='black', drawstyle='steps-post', legend=False)
    ax.axhline(y=0, color='grey', linestyle='--')
    ax.axhline(y=1, color='grey', linestyle='--')
    ax.set_ylabel('Fn(x)')
    return ax



'''
Created on Jun 30, 2020

@author: petergedeck
'''
from dataclasses import dataclass
from typing import List

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


@dataclass
class AcceptanceSamplingPlan:
    p: List[float]
    n: List[float]
    OC: List[float]
    ASN: List[float]
    AOQ: List[float]
    ATI: List[float]

    def plot(self, axes=None):
        if axes is None:
            _, axes = plt.subplots(nrows=2, ncols=2, figsize=[8, 8])

        samplePlan = self

        ax = axes[0, 0]
        ax.plot(self.p, samplePlan.OC)
        ax.set_xlabel('Fraction Nonconforming p')
        ax.set_ylabel('Probability of Acceptance')
        ax.set_title('OC Curve')

        ax = axes[0, 1]
        if self.ASN is not None:
            ax.plot(self.p, samplePlan.ASN)
            ax.set_xlabel('Fraction Nonconforming p')
            ax.set_ylabel('Average sample size')
            ax.set_title(f'maximum ASN = {max(samplePlan.ASN):.0f}')

        if self.n is not None:
            ax.plot(self.p, self.n)
            ax.set_xlabel('Fraction Nonconforming p')
            ax.set_ylabel('Uncurtailed sample size')

        ax = axes[1, 0]
        ax.plot(self.p, samplePlan.AOQ)
        ax.set_xlabel('Fraction Nonconforming p')
        ax.set_ylabel('AOQ')
        ax.set_title(f'AOQL = {max(samplePlan.AOQ):.4g}')

        if self.ATI is not None:
            ax = axes[1, 1]
            ax.plot(self.p, self.ATI)
            ax.set_xlabel('Fraction Nonconforming p')
            ax.set_ylabel('ATI')
        else:
            axes[1, 1].set_axis_off()


@dataclass
class SequentialSamplePlan:
    # R: dodge.SeqDesignBinomial
    AQL: float  # Acceptable quality level
    alpha: float  # alpha producer’s risk
    LQL: float  # LQL Limiting quality level
    beta: float  # beta consumers’ risk

    h1: float
    h2: float
    s: float
    h: List[float]
    p: List[float]

    k: List[int]
    accept: List[float]
    reject: List[float]

    N: int = None  # lot size, required for hypergeometric
    oc_type: str = 'binomial'

    def __repr__(self):
        return pd.Series({'h1': self.h1, 'h2': self.h2, 's': self.s})

    def __post_init__(self):
        k1 = ((1 - self.beta) / self.alpha) ** self.h
        k2 = (self.beta / (1 - self.alpha)) ** self.h
        self.OC = (k1 - 1) / (k1 - k2)
        self.AOQ = self.p * self.OC

        k5 = self.OC * np.log(self.beta / (1 - self.alpha))
        k3 = k5 + (1 - self.OC) * np.log((1 - self.beta) / self.alpha)
        k4 = self.p * np.log(self.LQL / self.AQL) + (1 - self.p) * np.log((1 - self.LQL) / (1 - self.AQL))
        self.ASN = k3 / k4

        self.ATI = self.N and k5 / k4 + (1 - self.OC) * self.N

    def plot(self):
        asp = AcceptanceSamplingPlan(p=self.p, OC=self.OC, ASN=self.ASN, AOQ=self.AOQ, ATI=self.ATI, n=None)
        asp.plot()

    def acceptanceChart(self, ax=None):
        if ax is None:
            _, ax = plt.subplots(figsize=[5, 5])
        ax.plot(self.k, self.accept, color='C0')
        ax.plot(self.k, self.reject, color='C0')
        ax.grid(linestyle='--', color='lightgrey')

        ax.set_ylim(min(self.accept), max(self.reject))
        ax.set_xlabel('k')
        ax.set_ylabel(r'$d_k$')
        ax.set_title('Sequential Acceptance Chart')

        x = np.median(self.k)
        common = {'horizontalalignment': 'center', 'verticalalignment': 'center'}
        ax.text(x, 0.5 * (np.min(self.accept) + np.mean(self.accept)), "ACCEPT", **common)
        ax.text(x, 0.5 * (np.max(self.reject) + np.mean(self.reject)), "REJECT", **common)
        ax.text(x, max(self.accept), "CONTINUE", **common)



'''
Created on Jun 20, 2020

@author: petergedeck
'''
from scipy import linalg

from mistat.mqcc.statistics import mqccStatistics
from mistat.qcc.rules import shewhartRules
import matplotlib.pyplot as plt
import pandas as pd


class MultivariateQualityControlChart:
    def __init__(self, data, qcc_type=mqccStatistics.default,
                 center=None, cov=None,
                 limits=True, pred_limits=False, rules=None,
                 data_name=None, labels=None,
                 newdata=None, newlabels=None,
                 confidence_level=None):
        self.statistic = mqccStatistics.get(qcc_type)
        self.qcc_type = self.statistic.qcc_type

        self.sizes = self.statistic.get_sizes(data)
        num_samples, sample_sizes, p = self.sizes
        if isinstance(num_samples, list):
            raise ValueError('varying number of samples (rows)')
        if isinstance(sample_sizes, list):
            raise ValueError('varying sample size (columns)')
        if sample_sizes == 1:
            self.statistic = mqccStatistics.get('T2single')
            self.qcc_type = self.statistic.qcc_type

        if confidence_level is None:
            confidence_level = (1 - 0.0027) ** p
        if not (0 < confidence_level < 1):
            raise ValueError('confidence.level must be a numeric value in the range (0,1)')

        if labels is None:
            labels = list(range(len(data)))
        self.labels = labels

        self.stats = self.statistic.stats(data, center=center, cov=cov)

        if newdata is not None:
            self.newdata = newdata
            self.newsizes = self.statistic.get_sizes(newdata)
            self.newstats = self.statistic.stats(newdata, center=self.stats.center, cov=self.stats.cov)
            self.newlabels = newlabels
            if newlabels is None:
                self.newlabels = list(range(len(data), len(data) + len(newdata)))
            # gives self.newdata, self.newstats, self.newsizes
            pred_limits = True
        else:
            self.newdata = None
            self.newstats = None
            self.newsizes = None

        if isinstance(limits, bool):
            if limits:
                limits = self.statistic.limits(num_samples, sample_sizes, p, confidence_level)
                self.limits = limits['control']
            else:
                self.limits = None
        else:
            self.limits = limits

        if isinstance(pred_limits, bool):
            if pred_limits:
                pred_limits = self.statistic.limits(num_samples, sample_sizes, p, confidence_level)
                self.pred_limits = pred_limits['prediction']
            else:
                self.pred_limits = None
        else:
            self.pred_limits = pred_limits

        self.violations = shewhartRules(self, run_length=0)
        self.violations['beyondPredLimits'] = shewhartRules(
            self, run_length=0, limits=self.limits)['beyondLimits']

    def plot(self, title=None, ax=None, show_legend=True):
        if ax is None:
            _, ax = plt.subplots(figsize=(8, 6))
        beyondLimits = [*self.violations['beyondLimits']['LCL'], *self.violations['beyondLimits']['UCL']]
        violatingRuns = self.violations['violatingRuns']
        df = pd.DataFrame({'x': self.labels, 'y': self.stats.statistics})
        ax = df.plot.line(x='x', y='y', style='-o', color='lightgrey',
                          marker='o', markerfacecolor='black', ax=ax)
        if self.newdata is not None:
            new_df = pd.DataFrame({'x': self.newlabels, 'y': self.newstats.statistics})
            ax = new_df.plot.line(x='x', y='y', style='-o', color='lightgrey',
                                  marker='o', markerfacecolor='black', ax=ax)
            df = pd.concat([df, new_df])
            ax.axvline(0.5 * (self.newlabels[0] + self.labels[-1]), color='black', linestyle='--')

        ax.plot(df.iloc[beyondLimits]['x'], df.iloc[beyondLimits]['y'], linestyle='None',
                marker='s', markerfacecolor='red', markeredgecolor='red')
        ax.plot(df.iloc[violatingRuns]['x'], df.iloc[violatingRuns]['y'],
                linestyle='None', marker='s', markerfacecolor='red', markeredgecolor='red')

        ax.legend().remove()
        ax.set_xlabel('Group')
        ax.set_ylabel('Group summary statistics')

        yticks = {'ticks': [], 'labels': []}
        if self.limits is not None and len(self.limits) == 1:
            ax.axhline(self.limits.LCL[0], color='black', linestyle='--')
            ax.axhline(self.limits.UCL[0], color='black', linestyle='--')

            yticks['ticks'].extend([self.limits.LCL[0], self.limits.UCL[0]])
            yticks['labels'].extend(['LCL', 'UCL'])

        if self.pred_limits is not None and len(self.pred_limits) == 1:
            ax.axhline(self.pred_limits.LPL[0], color='black', linestyle=':')
            ax.axhline(self.pred_limits.UPL[0], color='black', linestyle=':')

            if yticks['ticks'] and yticks['ticks'][0] == self.pred_limits.LPL[0]:
                yticks['labels'][0] = 'LCL/LPL'
            else:
                yticks['labels'].append('LPL')
                yticks['ticks'].append(self.pred_limits.LPL[0])

            yticks['labels'].append('UPL')
            yticks['ticks'].append(self.pred_limits.UPL[0])

        if yticks['ticks']:
            ax2 = ax.twinx()
            ax2.set_ylim(*ax.get_ylim())
            ax2.yaxis.tick_right()
            ax2.set_yticks(yticks['ticks'])
            ax2.set_yticklabels(yticks['labels'])

        fig = ax.get_figure()
        if show_legend:
            fig.subplots_adjust(bottom=0.2)
            fig.text(0.1, 0.1, f'Number of groups = {len(self.labels)}', fontsize=12)
            fig.text(0.1, 0.06, f'Sample size = {self.sizes.samples_sizes}', fontsize=12)
            fig.text(0.1, 0.02, f'|S| = {linalg.det(self.stats.cov):.5g}', fontsize=12)
            if self.limits is not None and len(self.limits) == 1:
                fig.text(0.4, 0.1, f'LCL = {self.limits.LCL[0]:.5g}', fontsize=12)
                fig.text(0.4, 0.06, f'UCL = {self.limits.UCL[0]:.5g}', fontsize=12)
            if self.pred_limits is not None and len(self.pred_limits) == 1:
                fig.text(0.6, 0.1, f'LPL = {self.pred_limits.LPL[0]:.5g}', fontsize=12)
                fig.text(0.6, 0.06, f'UPL = {self.pred_limits.UPL[0]:.5g}', fontsize=12)
            fig.text(0.4, 0.02, f'Number beyond limits = {len(beyondLimits)}', fontsize=12)

        if self.newdata is not None:
            ymax = max(*self.stats.statistics, *self.newstats.statistics)
            ymin = min(*self.stats.statistics, *self.newstats.statistics)
            ymax = 1.05 * ymax - 0.05 * ymin
            xcalibration = 0.5 * (self.labels[0] + self.labels[-1])
            xnewdata = 0.5 * (self.newlabels[0] + self.newlabels[-1])
            ax.text(xcalibration, ymax, 'Calibration data', ha='center', va='top')
            ax.text(xnewdata, ymax, 'New data', ha='center', va='top')

        fig_title = [f'{self.statistic.qcc_type} Chart']
        if title is not None:
            fig_title.append(title)
        fig.suptitle('\n'.join(fig_title), fontsize=14)
        return ax



'''
Created on Jul 6, 2020

@author: petergedeck
'''
from collections import namedtuple

from scipy import stats

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from .statistics import qccStatistics, Base_statistic


class Cusum:
    def __init__(self, data, center=None, std_dev=None, sizes=None, head_start=0, decision_interval=5,
                 se_shift=1, newdata=None):
        data = pd.DataFrame(data)
        if sizes is None:
            sizes = Base_statistic.getSizes(data)
        elif isinstance(sizes, int):
            sizes = np.array([sizes] * len(data))
        elif len(sizes) != len(data):
            raise ValueError('sizes length doesn'' match with data')
        if decision_interval <= 0:
            raise ValueError('decision_interval must be positive')

        if not (0 <= head_start < decision_interval):
            raise ValueError('head_start must be non-negative and less than decision_interval')

        qcc_type = 'xbarone' if any(size == 1 for size in sizes) else 'xbar'
        if len(data.columns) == 1 and any(size > 1 for size in sizes) and std_dev is None:
            raise ValueError('sizes larger than 1 but data appears to be single samples. '
                             'In this case you must provide also the std_dev')

        qccStatistic = qccStatistics.get(qcc_type)
        statistics = qccStatistic.stats(data, sizes)
        center = center
        if center is None:
            center = statistics.center
        std_dev = qccStatistic.sd(data, std_dev, sizes=sizes)

        self.newdata = None
        if newdata is not None:
            raise NotImplementedError()

        z = (statistics.statistics.flatten() - center) / (std_dev / np.sqrt(sizes))
        ldb = -decision_interval
        udb = decision_interval

        z_f = z - se_shift / 2
        cusum_pos = [max(0, head_start + z_f[0])]
        for zfi in z_f[1:]:
            cusum_pos.append(max(0, cusum_pos[-1] + zfi))
        cusum_pos = np.array(cusum_pos)

        z_f = z + se_shift / 2
        cusum_neg = [max(0, head_start - z_f[0])]
        for zfi in z_f[1:]:
            cusum_neg.append(max(0, cusum_neg[-1] - zfi))
        cusum_neg = np.array([-v for v in cusum_neg])

        violations = {'lower': np.where(cusum_neg < ldb)[0],
                      'upper': np.where(cusum_pos > udb)[0]}

        self.data = data
        self.statistics = statistics.statistics
        self.sizes = sizes
        self.center = center
        self.std_dev = std_dev

        self.pos = cusum_pos
        self.neg = cusum_neg
        self.head_start = head_start
        self.decision_interval = decision_interval
        self.se_shift = se_shift
        self.ldb = ldb
        self.udb = udb
        self.violations = violations

    def plot(self, ax=None, title='cusum Chart', xlabel='Group', ylabel='Cumulative Sum'):
        if ax is None:
            _, ax = plt.subplots(figsize=(8, 6))

        if self.newdata is not None:
            raise NotImplementedError()
        else:
            statistics = self.statistics
            indices = list(range(len(statistics)))

        if title is not None:
            ax.set_title(title)

        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)

        ax.axhline(0, linewidth=2, color='grey')
        ax.axhline(self.decision_interval, linestyle=':', color='grey')
        ax.axhline(-self.decision_interval, linestyle=':', color='grey')

        indices = np.array(indices)
        for cs, v in [(self.pos, self.violations['upper']), (self.neg, self.violations['lower'])]:
            v = np.array(v)
            ax.plot(indices, cs, color='grey')
            ax.plot(indices[v], cs[v], marker="s", color='red')
            ax.plot(np.delete(indices, v), np.delete(cs, v), color='black', marker='o', markersize=4)

        ax2 = ax.twinx()
        ax2.set_ylim(*ax.get_ylim())
        ax2.yaxis.tick_right()
        ax2.set_yticks([self.ldb, self.udb])
        ax2.set_yticklabels(['LDB', 'UDB'])

        fig = ax.get_figure()
        fig.subplots_adjust(bottom=0.23)
        fig.text(0.2, 0.1, f'Number of groups = {len(self.statistics)}', fontsize=12)
        fig.text(0.2, 0.06, f'Center = {self.center:.5g}', fontsize=12)
        fig.text(0.2, 0.02, f'StdDev = {self.std_dev:.5g}', fontsize=12)

        fig.text(0.5, 0.1, f'Decision interval (std. err.) = {self.decision_interval}', fontsize=12)
        fig.text(0.5, 0.06, f'Shift detection (std. err.) {self.se_shift}', fontsize=12)
        fig.text(
            0.5, 0.02, f"No. of points beyond limits = {len(self.violations['upper']) + len(self.violations['lower'])}", fontsize=12)


def cusumArl(*, randFunc=None, N=1000, limit=10_000, seed=None,
             kp=1, km=-1, hp=3, hm=-3,
             side='both', verbose=False):
    side = side.lower()
    if side not in ("both", "upper", "lower"):
        raise ValueError("side = '{side}' is not supported.")
    np.random.seed(seed)

    randFunc = randFunc or stats.norm()

    result = {}
    result['run'] = [runLength(randFunc.rvs(limit), kp, km, hp, hm, side) for _ in range(N)]
    rls = np.array([run.rl for run in result['run']])
    result['rls'] = rls
    # Ignore inf values for statistics calculations
    rls = np.ma.masked_invalid(rls)
    result['statistic'] = {
        'ARL': np.mean(rls),
        'Std. Error': np.sqrt((np.mean(rls ** 2) - np.mean(rls)) / N),
    }
    if verbose:
        stats = result['statistic']
        print(f"ARL {stats['ARL']:.5g}  Std. Error {stats['Std. Error']:.5g}")
    return result


def cusumPfaCed(*, randFunc1=None, randFunc2=None,
                tau=10, N=100, limit=10_000, seed=None,
                kp=1, km=-1, hp=3, hm=-3,
                side='both', verbose=True):
    side = side.lower()
    if side not in ("both", "upper", "lower"):
        raise ValueError("side = '{side}' is not supported.")
    np.random.seed(seed)

    rls = []
    for _ in range(N):
        x = [*randFunc1.rvs(tau), *randFunc2.rvs(limit - tau)]
        rls.append(runLength(x, kp, km, hp, hm, side).rl)
    result = {'rls': rls}

    rls = np.ma.masked_invalid(rls)
    pfa = np.mean(rls < tau)
    ced = np.mean(rls[rls >= tau]) - tau
    se = np.sqrt(
        (np.sum(rls[rls >= tau] ** 2) / (N - np.sum(rls < tau)) - ced ** 2)
        /
        (N - np.sum(rls < tau)))

    result['statistic'] = {
        'PFA': pfa,
        'CED': ced,
        'Std. Error': se
    }
    if verbose:
        stats = result['statistic']
        print(f"PFA {stats['PFA']:.5g}  CED {stats['CED']:.5g}  Std. Error {stats['Std. Error']:.5g}")
    return result


#
#   res <- list(run = apply(data, MARGIN=1, .runLength, kp=kp, km=km, ubd=hp, lbd=hm, side=side))
#   res$rls <- sapply(res$run, function(x) x$rl)
#
#   pfa <- sum(res$rls < tau)/nrow(data)
#   ced = sum(res$rls[res$rls >= tau])/(nrow(data) - sum(res$rls < tau)) - tau
#   se = sqrt((sum(res$rls[res$rls >= tau]^2)/(nrow(data) - sum(res$rls < tau)) - ced ^ 2) / (nrow(data) - sum(res$rls < tau)))
#
#   res$statistic <- c(pfa, ced, se)
#   names(res$statistic) <- c("PFA", "CED", "Std. Error")
#
#   if(printSummary)
#     print(res$statistic)
#
#   invisible(res)


def runLength(x, kp, km, ubd, lbd, side):
    x = np.array(x)

    if side in ['both', 'upper']:
        cusum = 0
        vUpper = np.inf
        for i, zfi in enumerate(x - kp):
            cusum = max(0, cusum + zfi)
            if cusum > ubd:
                vUpper = i
                break

    if side in ['both', 'lower']:
        cusum = 0
        vLower = np.inf
        for i, zfi in enumerate(x - km):
            cusum = min(0, cusum + zfi)
            if cusum < lbd:
                vLower = i
                break
    Result = namedtuple('RunLength', 'violationsLower,violationsUpper,rl')
    rl = 1
    if side == 'both':
        rl = min(vLower, vUpper)
    elif side == 'upper':
        rl = min(vUpper)
    elif side == 'lower':
        rl = min(vLower)
    return Result(vLower, vUpper, rl)



from pathlib import Path

DEMO_FILE_DIR = Path(__file__).parent

def globalWarmingBlogs():
    return {
        'blog-1': (DEMO_FILE_DIR / 'global warming blog 1.txt').read_text(),
        'blog-2': (DEMO_FILE_DIR / 'global warming blog 2.txt').read_text(),
    }



'''
Created on Jun 19, 2020

@author: petergedeck
'''
import pandas as pd
import pingouin as pg
import random


def randomizationTest(a, b, sample_stat, aggregate_stats=None, n_boot=500, seed=None,
                      printSummary=True):
    sample_values = pd.Series([*a, *b])
    sampleLabels = [*['a'] * len(a), *['b'] * len(b)]

    def apply_func(labels):
        stats = list(sample_values.groupby(labels).apply(sample_stat))
        if aggregate_stats is not None:
            stats = aggregate_stats(stats)
        return stats
    Bt0 = apply_func(sampleLabels)

    def apply_func_RSWOR(labels):
        # replace bootstrapped sample with RSWOR
        labels = random.sample(sampleLabels, len(sampleLabels))

        stats = list(sample_values.groupby(labels).apply(sample_stat))
        if aggregate_stats is not None:
            stats = aggregate_stats(stats)
        return stats
    random.seed(seed)
    B = pg.compute_bootci(sampleLabels, func=apply_func_RSWOR, n_boot=n_boot,
                          confidence=0.95, seed=seed, return_dist=True)

    if printSummary:
        # include the original stat in the quantile calculation
        orgPosition = sum(B[1] < Bt0) + 1
        quantile = 100 * orgPosition / (n_boot + 1)
        print(f'Original stat is {Bt0:.6f}')
        print(f'Original stat is at quantile {orgPosition} of {n_boot + 1} ' +
              f'({quantile:.2f}%)')
        dist = pd.Series(B[1]).describe()
        print('Distribution of bootstrap samples:')
        print(f" min: {dist['min']:.2f}, median: {dist['50%']:.2f},  max: {dist['max']:.2f}")
    return [*B[1], Bt0]



'''
Utility functions for "Data Mining for Business Analytics: Concepts, Techniques, and 
Applications in Python"

(c) 2019 Galit Shmueli, Peter C. Bruce, Peter Gedeck 
'''
from pathlib import Path
import unittest

import pytest

from mistat.data import DATA_DIR, get_description_file
import mistat
import pandas as pd


class TestData(unittest.TestCase):
    def test_load_data(self):
        with pytest.raises(ValueError):
            mistat.load_data('unknown data file')

        for name in ('CYCLT.csv.gz', 'CYCLT.csv', 'CYCLT'):
            data = mistat.load_data(name)
            assert isinstance(data, pd.Series)

    def test_load_data_all(self):
        for name in Path(DATA_DIR).glob('*.csv.gz'):
            data = mistat.load_data(name.name)
            assert isinstance(data, (pd.Series, pd.DataFrame))
            assert len(data.shape) <= 2
            if len(data.shape) == 1:
                assert isinstance(data, pd.Series)
            else:
                assert isinstance(data, pd.DataFrame)
                assert data.shape[1] > 1

    def _test_data_description_exists(self):
        for name in Path(DATA_DIR).glob('*.Rd'):
            data_file = name.with_suffix('.csv.gz')
            assert data_file.exists()
        for name in Path(DATA_DIR).glob('*.csv.gz'):
            description_file = get_description_file(name.name)
            assert description_file.exists()

    def _test_describe_data(self):
        with pytest.raises(ValueError):
            mistat.describe_data('unknown data file')

        for name in ('CYCLT.csv.gz', 'CYCLT.csv', 'CYCLT'):
            text = mistat.describe_data(name)
            assert isinstance(text, str)
            assert 'Description' in text

        for name in Path(DATA_DIR).glob('*.csv.gz'):
            text = mistat.describe_data(name.name)
            assert isinstance(text, str)
            assert 'Description' in text

    def _test_prepare_Rd_HTML(self):
        for name in Path(DATA_DIR / 'Rd').glob('*.Rd'):
            name = name.with_suffix('').name
            print(f'Rd2HTML(file("{name}.Rd"), "{name}.html")')
        for name in Path(DATA_DIR / 'Rd').glob('*.Rd'):
            name = name.with_suffix('').name
            print(f'pandoc {name}.html -t rst -o {name}.rst')
        for name in Path(DATA_DIR / 'Rd').glob('*.Rd'):
            name = name.with_suffix('').name
            print(f'pandoc {name}.rst -t markdown -o {name}.md')



'''
Utility functions for "Data Mining for Business Analytics: Concepts, Techniques, and 
Applications in Python"

(c) 2019 Galit Shmueli, Peter C. Bruce, Peter Gedeck 
'''
import unittest

import pytest

from mistat.data import load_data
from mistat.mqcc.multivariateQualityControlChart import MultivariateQualityControlChart


class TestMultivariateQualityControlChart(unittest.TestCase):
    def test_T2single_chart(self):
        almpin = load_data('ALMPIN')
        base = almpin.iloc[:30, ]
        almpinSubset = almpin.iloc[30:, ]

        center = base.mean()
        cov = base.cov()

        mqcc = MultivariateQualityControlChart(almpinSubset, qcc_type='T2single',
                                               center=center, cov=cov)



'''
Created on Jul 5, 2020

@author: petergedeck
'''
from collections import namedtuple

from scipy import stats

import numpy as np


def runsTest(sequence, cutoff=None, alternative=None, verbose=False):
    if alternative is None:
        alternative = 'two.sided'

    if alternative not in ['two.sided', 'less', 'greater']:
        raise ValueError("specify alternative as 'two.sided', 'less', or 'greater'")
    # convert sequence to binary if cutoff is given
    x = np.array(sequence)
    if cutoff is not None:
        x = np.array([0 if xi < cutoff else 1 for xi in x])

    if len(set(x)) != 2:
        raise ValueError('sequence must contain only two values or must be converted to binary sequence using cutoff')
    if cutoff is None:
        cutoff = np.mean([np.min(x), np.max(x)])

    n1 = np.sum(np.min(x) == x)
    n2 = np.sum(np.max(x) == x)
    runs = 1 + np.sum(x[1:] != x[:-1])

    runs_exp = 1 + (2 * n1 * n2) / (n1 + n2)
    std = np.sqrt(2 * n1 * n2 * (2 * n1 * n2 - n1 - n2) /
                  ((n1 + n2)**2 * (n1 + n2 - 1)))

    Z = (runs - runs_exp) / std

    if alternative == 'two.sided':
        pval = 2 * stats.norm.cdf(-abs(Z))
    elif alternative == 'less':
        pval = stats.norm.cdf(Z)
    elif alternative == 'greater':
        pval = stats.norm.cdf(-Z)

    Result = namedtuple('Result', 'statistic,pval,method,alternative')
    result = Result(Z, pval, 'Runs Test', alternative)
    if verbose:
        print(f"""
{result.method}

Standard Normal = {result.statistic:.4f}, p-value = {result.pval:.4f}
alternative hypothesis: {result.alternative}""")
    return result



'''
Created on Jul 4, 2020

@author: petergedeck
'''
from itertools import combinations
from itertools import product

from scipy import stats
import patsy

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns


def calculateMainEffects(df, response, factors=None):
    if response not in df.columns:
        raise ValueError(f'Response {response} not found in data')
    factors = factors or [c for c in df.columns if c != response]
    mainEffects = []
    for column in factors:
        for level, g in sorted(df[response].groupby(df[column])):
            mainEffects.append({'factor': column, 'level': level, 'mean': g.mean()})
    return pd.DataFrame(mainEffects)


def calculateInteractions(df, response, factors=None):
    factors = factors or [c for c in df.columns if c != response]
    interactions = []
    for f1, f2 in product(factors, factors):
        for (l1, l2), g in df[response].groupby([df[f1], df[f2]]):
            interactions.append({'f1': f1, 'l1': l1, 'f2': f2, 'l2': l2, 'mean': g.mean()})
    return pd.DataFrame(interactions)


def mainEffectsPlot(df, response, factors=None, col_wrap=None, aspect=0.4, height=3, continuous=False):
    factors = factors or [c for c in df.columns if c != response]
    mainEffects = calculateMainEffects(df, response, factors)
    col_wrap = col_wrap or len(factors)

    def effectPlot(x, y, **kwargs):
        ax = plt.gca()
        ax.plot(x, y, marker='s')
        ax.axhline(kwargs['meanEffect'], color='grey', linestyle=':')
        x = sorted(x)
        d = 0.35 * (x[-1] - x[0])
        ax.set_xlim(x[0] - d, x[-1] + d)
        if not continuous:
            ax.set_xticks(x)

    dy = 0.15 * (mainEffects['mean'].max() - mainEffects['mean'].min())
    ylim = [mainEffects['mean'].min() - dy, mainEffects['mean'].max() + dy]
    meanEffect = df[response].mean()

    g = sns.FacetGrid(mainEffects, col='factor', col_wrap=col_wrap, height=height,
                      sharex=False, aspect=aspect, despine=True, ylim=ylim)
    g.map(effectPlot, 'level', 'mean', meanEffect=meanEffect).set_titles('')
    g.fig.subplots_adjust(wspace=.02)
    for idx, column in enumerate(factors):
        g.axes[idx].set_xlabel(column)


def interactionPlot(df, response, factors=None):
    factors = factors or [c for c in df.columns if c != response]
    interactions = calculateInteractions(df, response, factors)

    def setLimits(set_lim, values, f):
        v = sorted(set(values))
        d = f * (v[-1] - v[0])
        set_lim(v[0] - d, v[-1] + d)

    def effectPlot(x1, x2, y, **kwargs):
        x2values = sorted(set(x2), reverse=True)
        linestyle = ['-', '--', ':']
        color = ['black', 'red', 'blue']
        marker = ['s', '^', 'o']

        ax = plt.gca()
        setLimits(ax.set_xlim, x1, 0.35)
        ax.set_xticks(sorted(set(x1)))
        if len(x1) == len(set(x1)):
            for idx, x2i in enumerate(x2values):
                yi = (idx + 1) / (len(x2values) + 1)
                i = idx % len(linestyle)
                ax.text(0.2, yi, f'{x2i:.4g}', transform=ax.transAxes, verticalalignment='center')
                ax.plot(0.1, yi, marker=marker[i], color=color[i], linestyle=linestyle[i], transform=ax.transAxes)
            return
        for x2i, group in pd.DataFrame({'x1': x1, 'x2': x2, 'y': y}).groupby('x2'):
            i = x2values.index(x2i) % len(linestyle)
            group.plot.line(x='x1', y='y', ax=ax,
                            marker=marker[i], color=color[i], linestyle=linestyle[i])

    dy = 0.15 * (interactions['mean'].max() - interactions['mean'].min())
    ylim = [interactions['mean'].min() - dy, interactions['mean'].max() + dy]

    g = sns.FacetGrid(interactions, col='f1', row='f2', sharex=False, despine=True, ylim=ylim, height=1.5)
    g.map(effectPlot, 'l1', 'l2', 'mean').set_titles('')
    g.fig.subplots_adjust(wspace=.05, hspace=.05)
    for idx, column in enumerate(factors):
        g.axes[-1][idx].set_xlabel(column)
        g.axes[idx][0].set_ylabel('')
        ax = g.axes[idx][idx]
        ax.text(0.9, 0.5, column, transform=ax.transAxes, verticalalignment='center', horizontalalignment='right')
    return ax


def getModelMatrix(design, mod=0, maxscale=1):
    ''' Convert design to model matrix. Will rescale design to code values '''
    design = design - design.mean()
    design = maxscale * design / design.max()

    kvar = design.shape[1]
    names = [f'x{i + 1}' for i in range(kvar)]
    design.columns = names

    formula = list(names)
    if mod >= 1:
        formula.extend(f'{f1}:{f2}' for f1, f2 in combinations(names, 2))
    if mod == 2:
        formula.extend(f'np.power({f1}, 2)' for f1 in names)

    dm = patsy.dmatrix(f'~ {"+".join(formula)}', data=design, return_type='dataframe')  # @UndefinedVariable

    def cleanColumnName(c):
        if not c.startswith('np.power'):
            return c
        c = c.split('(')[1].split(',')[0]
        return f'{c}**2'
    dm.columns = [cleanColumnName(c) for c in dm.columns]
    return dm


def FDS_Plot(design, mod=0, ax=None, plotkw=None, label='y', maxscale=1):
    ''' Fraction of design space plot '''
    dm = getModelMatrix(design, mod=mod, maxscale=maxscale)
#     design = design - design.mean()
#     design = maxscale * design / design.max()
#     kvar = design.shape[1]
#     if not (1 < kvar < 8):
#         raise ValueError('The design matrix must have between 2 and 7 variables')
#     names = [f'x{i + 1}' for i in range(kvar)]
#     design.columns = names
#
#     formula = list(names)
#     if mod >= 1:
#         formula.extend(f'{f1}:{f2}' for f1, f2 in combinations(names, 2))
#     if mod == 2:
#         formula.extend(f'np.power({f1}, 2)' for f1 in names)
#
#     dm = patsy.dmatrix(f'~ {"+".join(formula)}', data=design, return_type='dataframe')  # @UndefinedVariable

    terms = dm.columns
    kvar = design.shape[1]
    names = [f'x{i + 1}' for i in range(kvar)]

    XtX = np.matmul(np.transpose(dm).values, dm.values)
    XtXI = pd.DataFrame(np.linalg.inv(XtX),
                        index=terms, columns=terms)

    fX = {'Int': [1] * 5000}
    fX.update({f: stats.uniform.rvs(loc=-1, scale=2, size=5000) for f in names})
    if mod >= 1:
        for f1, f2 in combinations(names, 2):
            fX[f'{f1}:{f2}'] = fX[f1] * fX[f2]
    if mod == 2:
        for f in names:
            fX[f'np.power({f}, 2)'] = fX[f] * fX[f]
    fX = pd.DataFrame(fX)
    p1 = np.matmul(fX, XtXI)
    v = np.diag(np.matmul(p1.values, np.transpose(fX.values)))
    vi = np.sort(v)

    # calculate fraction of design space
    fs = np.arange(1, len(vi) + 1) / len(vi)
    plotkw = plotkw or {}
    plotkw['color'] = plotkw.get('color', 'black')
    # plotkw['legend'] = False
    ax = pd.DataFrame({'x': fs, label: vi}).plot(x='x', y=label, ax=ax, **plotkw)
    ax.axvline(0.5, linestyle=':', color='grey')
    ax.axhline((vi[2499] + vi[2500]) / 2, linestyle=':', color='grey')
    ax.set_xlabel('Fraction of Space')
    ax.set_ylabel('Relative Prediction Variance')
    return ax
